#!/usr/bin/python
"""
The WebFrame system.
This is the main module, providing generic stuff to mostly
the other modules and the user application.

Every app will want one of these.
"""
__author__ = "Bartek Kania <mrbk@gnarf.org>"

import codecs
import os
import logging
import wf
from WebFrame.options import opt_space
from WebFrame.session import Session
from WebFrame import template
from WebFrame import db
from WebFrame import auth
from WebFrame import reqresp


# The main data handle, for cached connections
__dbh = None

class Context(object):
    """Used to keep track of a request context.
    This object is passed around to all handlers and contains all one
    should want to know about the request"""
    
    # Hack to get them into the documentation properly
    req = property(fget = lambda self: self.__the_req,
                   doc = "The Request object for this request")
    resp = property(fget = lambda self: self.__the_resp,
                    doc = "The Response object for this request")
    sess = property(fget = lambda self: self.__the_sess,
                    doc = "The Session object for this request")
    template = property(fget = lambda self: self.__the_template,
                        doc = """The Template() settings object for this
                        request. See WebFrame.template for docs""")
    user = property(fget = lambda self: self.__the_user,
                    doc = "The currently logged in user. None if none")
    
    def __init__(self, req, resp):
        self.__the_req = req
        self.__the_resp = resp
        self.__the_template = template.Template(self)
        self.__the_user = None
        self.__the_sess = None
        self.set_lang(get_opt("lang.default"))

    def _set_sess(self, sess):
        self.__the_sess = sess

    def _set_user(self, user):
        self.__the_user = user
        if self.__the_sess != None:
            self.__the_sess["user"] = user

    def _clear_user(self):
        self.__the_user = None
        if self.__the_sess != None and "user" in self.__the_sess:
            del self.__the_sess["user"]

    def set_lang(self, lang):
        """Set the language for this context.
        Use ISO-693-* codes for clarity"""
        self.lang = lang
        
## end class Context()

def _init(app, logger):
    """Init-function for the main WebFrame system
    This is called from the web server when the WebFrame needs to be
    initialized, and this in turn initializes the app by importing the
    module refered to with PythonOption APPLICATION in the apache
    config.
    DO NOT CALL THIS YOURSELF!
    """
    global opt_space

    # Set up logging
    if logger:
        level = get_opt("logging.level")
        if level == None:
            if get_opt("debug"): level = logging.DEBUG
            else: level = logging.INFO
        rl = logging.getLogger('')
        rl.setLevel(level)
        logger.setFormatter(logging.Formatter(get_opt("logging.format")))
        rl.addHandler(logger)
    else:
        logging.basicConfig(level=level,
                            format=get_opt("logging.format"))
    # Import app
    opt_space["app.name"] = app
    __import__(app)

    # Setup subsystems
    cache.__connect(True)
    auth._init()


def open_db():
    """Open the default database.
    See pydoc WebFrame.db for details."""
    global __dbh
    if __dbh == None:
        __dbh = db.DB(get_opt("db.host"),
                      get_opt("db.name"),
                      get_opt("db.user"),
                      get_opt("db.pass"))
    return __dbh

def get_opt(opt):
    """Get a configuration option from the system.
    See pydoc WebFrame.options for option documentation"""
    global opt_space
    if opt in opt_space:
        return opt_space[opt]
    else:
        return None

def set_opt(opt, value):
    """Set a configuration option in the system.
    See pydoc WebFrame.options for option documentation"""
    global opt_space
    opt_space[opt] = value


def __init_processing(ctx):
    sess = Session(ctx)
    ctx._set_sess(sess)
    ctx._set_sess = None  # Noone may change session!
    sess._load()

    # Set user in session an context. If a user is present in the context,
    # via basic auth probably, then that takes precedence over user in session.
    if ctx.user != None:
        sess["user"] = ctx.user
    elif "user" in sess:
        ctx._set_user(sess["user"])

    # Check for query handlers
    h = handler.find_query(ctx.req.params)
    if h != None:
        h.run(ctx)
        
def __finish_processing(ctx):
    ctx.sess._save()

def _process_template(ctx):
    """Called from the web server to process a template"""
    try:
        __init_processing(ctx)
    except reqresp.StopProcessing:
        __finish_processing(ctx)
        return
    
    if get_opt("debug"):
        logging.debug("Template handler running for %s", ctx.req.filename)
    ctx.resp.content_type = get_opt("http.default_content_type")

    # Check path info
    if get_opt("debug"):
        logging.debug("Template path_info: %s", ctx.req.path_info)
    if not get_opt("template.allow_pathinfo") and ctx.req.path_info:
        return ctx.resp.not_found()

    # Check if the html-file is there and readable
    if os.access(ctx.req.filename, os.R_OK):
        ctx.resp.write(template.run_template(ctx, ctx.req.filename))
        __finish_processing(ctx)
        return
    else:
        logging.error("Template '%s' not found", ctx.req.filename)
        __finish_processing(ctx)
        return ctx.resp.not_found()

def _process_wf(ctx):
    """Called from the webserver to process a .wf file"""
    try:
        __init_processing(ctx)
    except reqresp.StopProcessing:
        __finish_processing(ctx)
        return

    if get_opt("debug"):
        logging.debug("WF handler running for %s", ctx.req.filename)
    try:
        st = wf.process(ctx.req.filename, ctx)
    except:
        __finish_processing(ctx)
        raise  # Re-raise as we only catch it to allow processing to finish
    __finish_processing(ctx)
    return st

def _process_handler(ctx, h):
    """Called from the webserver to process a software defined handler"""
    try:
        __init_processing(ctx)
    except reqresp.StopProcessing:
        __finish_processing(ctx)
        return

    if get_opt("debug"):
        logging.debug("Handler handler running for %s", ctx.req.uri)
    try:
        st = h.run(ctx)
    except Exception, e:
        __finish_processing(ctx)
        raise
    __finish_processing(ctx)
    return st
    

def _run_plain(func, ctx):
    try:
        st = func(ctx)
    except Exception, e:
        __finish_processing(ctx)
        raise
    __finish_processing(ctx)
    return st

def _run_xml(func, ctx):
    etree = ctx.req.body_xml()
    try:
        ctype, res = func(ctx, etree)
    except Exception, e:
        __finish_processing(ctx)
        raise
    if res != None:
        ctx.resp.content_type = ctype
        ctx.resp.write_xml(res)
    __finish_processing(ctx)
    return ctx.resp.stop_processing()

def _run_json(func, ctx):
    jp = ctx.req.body_json()
    try:
        res = func(ctx, jp)
    except Exception, e:
        __finish_processing(ctx)
        raise
    if res:
        ctx.resp.content_type = get_opt("json.content_type")
        ctx.resp.write_json(res)
    __finish_processing(ctx)
    return ctx.resp.stop_processing()
